<?php

/**
 * Contains functions related to managing
 * Access Control Lists on an IMAP server using RFC 2086.
 *
 * $Horde: imp/lib/Acl/rfc2086.php,v 1.1  $
 *
 * Copyright 2003 Chris Hastie <imp@oak-wood.co.uk>
 *
 * See the enclosed file COPYING for license information (GPL). If you
 * did not receive this file, see http://www.fsf.org/copyleft/gpl.html.
 *
 * @author  Chris Hastie <imp@oak-wood.co.uk>
 * @version $Revision: 1.1 $
 * @since   IMP 4.0
 * @package imp
 */

class IMP_Acl_rfc2086 extends IMP_Acl {

    /* List of servers capabilities, output of 
     * CAPABILITY command. Formated as a hash
     * Array(
     *     *capability* => 1
     * )
     *
     * @var array $_caps
     */
    var $_caps = array();
    
    /**
     * Constructor.
     *
     * @access public
     *
     * @param array $params  Any additional parameters this driver may need.
     */
    function IMP_Acl_rfc2086($params = array())
    {
        $this->_params = $params;
        
        $this->_caps = $this->_getCapability();
        
        // if we couldn't get the server's capability, we'll assume
        // ACL is supported for now.
        if (!empty($this->_caps) && !array_key_exists('acl', $this->_caps)) {
            $this->isSupported = false;
        } else {
            $this->isSupported = true;
        }
        
        $this->protected = array($_SESSION['imp']['user']);
        
        $this->rightsList = array (
             'l' => _("Visible - user can see the folder"),
             'r' => _("Read messages"),
             's' => _("Mark with seen / unseen flags"),
             'w' => _("Mark with other flags (eg Important / Answered)"),
             'i' => _("Insert messages"),
             'p' => _("Post to this folder (not enforced by IMAP)"),
             'c' => _("Create sub folders"),
             'd' => _("Delete and purge messages"),
             'a' => _("Administer - set permissions for other users")
        );
        
        
    }

    /**
     * Sets the ACL on an IMAP server
     *
     * @access public
     *
     * @param string $folder  The folder on which to edit the ACL
     *                           
     * @param string $share_user  The user to grant rights to
     *
     * @param array $acl  An array, the keys of which are the 
     *                    rights to be granted (see RFC 2086)
     *
     * @return mixed  True on success, false on failure unless
     *                server doesn't support ACLs, returns 'no_support'
     */

    function createAcl($folder, $share_user, $acl) {

        global $imp, $notification;
        $acl_str = '';

        if (!empty($acl)) {
            foreach ($acl as $key => $val) {
                $acl_str .= $key;
            }
        }
        $result = imap_setacl($imp['stream'], $folder, $share_user, $acl_str);

        if (!$result) {
            $errors = @imap_errors();
            foreach ($errors as $err) {
                if ($err == "ACL not available on this IMAP server"){
                    $err = _("This IMAP server does not support sharing folders.");
                    $result = 'no_support';
                }
                $notification->push($err, 'horde.error');
            }
        } elseif (strlen($acl_str) <1) {
            $notification->push(_("All rights on folder ")."'".$folder."'"._(" successfully removed for user ")."'".$share_user."'" , 'horde.success');
        }
        else {
            $notification->push(_("User")." '".$share_user."' "._("successfully given the following rights for the folder ")."'".$folder."': ".$acl_str , 'horde.success');
        }

        return $result;

    }

    /**
     * Edits an ACL on an IMAP server
     *
     * @access public
     *
     * @param string $folder  The folder on which to edit the ACL
     *                           
     * @param string $share_user  The user to grant rights to
     *
     * @param array $acl  An array, the keys of which are the 
     *                    rights to be granted (see RFC 2086)
     *
     * @return mixed  True on success, false on failure unless
     *                server doesn't support ACLs, returns 'no_support'
     */

    function editAcl($folder, $share_user, $acl) {
        return $this->createAcl($folder, $share_user, $acl);
    }    

    /**
     * Attempts to get the result of a CAPABILITY command to 
     * the current IMAP server. If protocol is imap/ssl, will
     * only attempt ssl connection with PHP >= 4.3
     *
     * @access public
     *
     * @return array  An array containing the server's capabilities

     */

    function _getCapability() {
        global $notification;
        $capabilities = null;
        $sub_proto = null;

        if (($pos = strpos($_SESSION['imp']['protocol'], '/'))) {
            $sub_proto = substr($_SESSION['imp']['protocol'], $pos+1);
        }  

        $server = $_SESSION['imp']['server'];

        if ($sub_proto == "ssl") {
            if (version_compare( phpversion(), "4.3.0") >= 0 ){
                $server = $sub_proto."://".$server;
            } else {
                $notification->push(_("Could not retrieve server's capabilities"),'horde.warn');
                return null;
            }
        }

        $imap = fsockopen($server, $_SESSION['imp']['port'], $errno, $errstr, 30);

        if (!$imap) {
            $notification->push(_("Could not retrieve server's capabilities"),'horde.warn');
            return null;
        } else {
            $response = fgets($imap, 4096);
            if (preg_match("/^\*\sOK/", $response)) {
                fputs($imap, "x CAPABILITY\r\n");            
                $response =  trim(fgets ($imap,1024));
                if (preg_match("/^\*\sCAPABILITY/", $response)) {
                    $response_array = explode(" ", $response);
                    foreach ($response_array as $var){
                        if (preg_match("/=/", $var)) {
                            $var2 = explode("=", $var, 2);
                            $capabilities[strtolower($var2[0])][strtolower($var2[1])]=1;
                        } else {
                            $capabilities[strtolower($var)] = 1;
                        }
                    }

                } else {
                    $notification->push(_("Could not retrieve server's capabilities"),'horde.warn');
                }
            } else {
                $notification->push(_("Could not retrieve server's capabilities"),'horde.warn');
            }
            fclose ($imap);
        }   
        return $capabilities;
    }

    /**
     * Attempts to retrieve the existing ACL for a folder from 
     * the current IMAP server. 
     *
     * @access public
     *
     * @param string folder  The folder to get the ACL for
     *
     * @return array  A hash containing information on the ACL
     *                Array (
     *                    user => Array (
     *                                right => 1
     *                            )
     *                )

     */
    function getAcl ($folder) {
        if (!empty($this->_caps['auth'])) {
            if (array_key_exists('digest-md5', $this->_caps['auth'])){
                return $this->_getAcl($folder, 'digest-md5');
            } elseif (array_key_exists('cram-md5', $this->_caps['auth'])) {
                return $this->_getAcl($folder, 'cram-md5');
            } else {
                return $this->_getAcl($folder, 'login');
            }
        } else {
            return $this->_getAcl($folder, 'login');
        }
    }

    /**
     * Attempts to retrieve the existing ACL for a folder from 
     * the current IMAP server. If protocol is imap/ssl, will
     * only attempt ssl connection with PHP >= 4.3
     *
     * NB: if Auth_SASL is not installed this function will send
     * the users password to the IMAP server as plain text!!
     *
     * @access private
     *
     * @param string folder  The folder to get the ACL for
     *
     * @param string authMech  The authorisation mechanism to use
     *                         One of cram-md5, digest-md5 or login
     *
     * @return array  A hash containing information on the ACL
     *                Array (
     *                    user => Array (
     *                                right => 1
     *                            )
     *                )

     */
    function _getAcl ($folder, $authMech) {

        global $notification; 
        $returnAcl = array();
        $sub_proto = null;
        $txid = 0;    

        $have_sasl = false;
        // silence warnings during check if Auth_SASL module
        // is installed
        $old_err_level = error_reporting(0);    
        if (include_once 'Auth/SASL.php'){ 
            $have_sasl = true;
        }
        error_reporting($old_err_level);

        $pass = Secret::read(Secret::getKey('imp'), $_SESSION['imp']['pass']);

        if (($pos = strpos($_SESSION['imp']['protocol'], '/'))) {
            $sub_proto = substr($_SESSION['imp']['protocol'], $pos+1);
        }  

        $server = $_SESSION['imp']['server'];

        if ($sub_proto == "ssl") {
            if (version_compare( phpversion(), "4.3.0") >= 0 ){
                $server = $sub_proto."://".$server;
            } else {
                $notification->push(_("Could not retrieve ACL"),'horde.warn');
                return null;
            }
        }

        // Quote the folder string if it contains non alpha-numeric 
        // characters
        if (preg_match("/\W/",$folder)) {
            $folder = '"'.$folder.'"';
        } 

        $imap = fsockopen($server, $_SESSION['imp']['port'], $errno, $errstr, 30);

        if (!$imap) {
            $notification->push(_("Could not retrieve ACL"),'horde.warn');
            return null;
        } else {
            $response = fgets($imap, 4096);
            if (preg_match("/^\*\sOK/", $response)) {             

                // login using the preferred mechanism
                // default to login if Auth_SASL is not installed
                if (($authMech == 'cram-md5') && $have_sasl) {
                    $login = Auth_SASL::factory('crammd5');

                    fputs ($imap, "$txid AUTHENTICATE CRAM-MD5\r\n");  
                    $challenge = explode (" ",trim(fgets ($imap, 1024)));

                    $response = $login->getResponse($_SESSION['imp']['user'], $pass, base64_decode($challenge[1]));
                    fputs($imap, base64_encode($response)."\r\n");

                } elseif(($authMech == 'digest-md5') && $have_sasl) {
                    $login = Auth_SASL::factory('digestmd5');

                    fputs ($imap, "$txid AUTHENTICATE DIGEST-MD5\r\n");  
                    $challenge = explode (" ",trim(fgets ($imap, 1024)));

                    $response = $login->getResponse($_SESSION['imp']['user'], $pass, base64_decode($challenge[1]),
                        $_SESSION['imp']['server'], $_SESSION['imp']['base_protocol']);

                    fputs($imap, base64_encode($response)."\r\n");
                    $response = explode (" ", trim(fgets ($imap,1024)));
                    $response = base64_decode($response[1]);
                    if (!preg_match("/rspauth=/", $response)) {
                        $notification->push(_("Could not retrieve ACL"),'horde.warn');
                        return null;                    
                    }
                    fputs($imap, "\r\n");

                } else {
                    fputs ($imap, "$txid LOGIN ".$_SESSION['imp']['user']." ".$pass."\r\n");

                }
                $response = trim(fgets ($imap,1024));
                if (preg_match("/^$txid\sOK/", $response)) {
                    $txid++;
                    fputs ($imap, "$txid GETACL ".$folder."\r\n");
                    $response =  trim(fgets ($imap,4096));
                    if (preg_match ("/^\*\s+ACL\s+(.*)/i", $response, $matches)){                    
                        $res_arr = $this->_atomise($matches[1]);
                        $res_folder = array_shift($res_arr);
                        $is_key = 1;
                        $key = null;
                        foreach ($res_arr as $var){
                            if ($is_key){
                                $key = $var;
                                $is_key = 0;
                            } else {
                                $perms = preg_split('//', $var, -1, PREG_SPLIT_NO_EMPTY);
                                foreach ($perms as $p_key => $p_var){
                                    $returnAcl[$key][$p_var] = 1;
                                }                             
                                $is_key = 1;
                            }
                        }
                    } else {
                        $notification->push(_("Could not retrieve ACL"),'horde.warn');
                    }
                } else {
                    $notification->push(_("Could not retrieve ACL"),'horde.warn');
                }
            } else {
                $notification->push(_("Could not retrieve ACL"),'horde.warn');
            }
            fclose ($imap);
        }   
        return $returnAcl;
    }

    /**
     * Can a user edit this ACL? Returns true if $user has
     * the 'a' right, false otherwise
     *
     * @param array $acl  An array as returned by getAcl()
     *
     * @param string $user  A user name
     *
     * @returns boolean  True if $user has 'a' right or $acl is empty
     */
    
    
    function canEdit($acl, $user) {
        /* Possible for getAcl to fail but createAcl still to work 
           (eg ssl, PHP <4.3), so return true if $acl is empty */
        if (empty($acl)){
            return true;
        } elseif (array_key_exists('a', $acl[$user]) && $acl[$user]['a'] >0 ) {
            return true;
        } else {
            return false;
        }
    }
    
    /**
    * Crudely split a string into 'atoms'
    *
    * @param string $in  The string to split
    *
    * @returns array  An array of 'atoms'
    *
    */
    
    function _atomise($in) {
        $length = strlen($in);
        $qt = false;
        $idx = 0;
        $out = array();
        for ($i = 0; $i < $length; $i++) { 
            $char = substr($in, $i, 1);
            if ($char == '"' && !$qt) {
                $qt = true;
                $idx ++;
                continue;
            } elseif ($char == " " && !$qt) {
                $idx ++;
            } elseif ($char == '"' && $qt) {
                $qt = false;
                $idx ++;
            } else {
                if (empty($out[$idx])){
                    $out[$idx] = $char;
                } else {
                    $out[$idx] .= $char;
                }
            }
        }
        return $out;
    }
}    